from __future__ import annotations

from typing import TYPE_CHECKING

from mipac import Note
from mipac.models.lite.user import LiteUser

if TYPE_CHECKING:
    from mipa.ext.commands import CMD, BotBase


class Context:
    __slots__ = ('__message', 'bot', 'args', 'kwargs', 'command', '__cmd')

    def __init__(
        self, *, message, bot: BotBase, args: tuple | None=None, kwargs=None, cmd: CMD = None
    ):
        self.__message: Note = message
        self.bot: BotBase = bot
        self.args: tuple = args or ()
        self.kwargs = kwargs or {}
        self.command = cmd.func
        self.__cmd = cmd

    @property
    def message(self) -> Note:
        return self.__message

    @property
    def author(self) -> LiteUser:
        return self.__message.author

    @property
    def cog(self):
        return self.bot.get_cog(self.__cmd.cog_name)



from __future__ import annotations

import asyncio
import importlib
import inspect
import logging
import re
import sys
import traceback
from typing import Any, Callable, Coroutine, Dict, Optional, Tuple, Union

from aiohttp import ClientWebSocketResponse
from mipac.client import Client as API
from mipac.manager import ClientActions
from mipac.models.user import UserDetailed

from mipa.exception import WebSocketReconnect
from mipa.gateway import MisskeyWebSocket
from mipa.router import Router
from mipa.state import ConnectionState
from mipa.utils import LOGING_LEVEL_TYPE, setup_logging

_log = logging.getLogger()


class Client:
    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        max_capture: int = 100,
        **options: Dict[Any, Any],
    ):
        super().__init__(**options)
        self.max_capture = max_capture
        self._router: Router
        self.url = None
        self.extra_events: Dict[str, Any] = {}
        self.special_events: Dict[str, Any] = {}
        self.token: Optional[str] = None
        self.origin_uri: Optional[str] = None
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self.core: API
        self._connection: ConnectionState
        self.user: UserDetailed
        self.ws: MisskeyWebSocket

    def _get_state(self, **options: Any) -> ConnectionState:
        return ConnectionState(
            dispatch=self.dispatch, loop=self.loop, client=self
        )

    async def on_ready(self, ws: ClientWebSocketResponse):
        """
        on_readyのデフォルト処理

        Parameters
        ----------
        ws : WebSocketClientProtocol
        """

    def event(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_event(func, name)
            return func

        return decorator

    def add_event(
        self, func: Coroutine[Any, Any, Any], name: Optional[str] = None
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.special_events[name].append(func)
        else:
            self.special_events[name] = [func]

    def listen(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_listener(func, name)
            return func

        return decorator

    def add_listener(
        self,
        func: Union[Coroutine[Any, Any, Any], Callable[..., Any]],
        name: Optional[str] = None,
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')
        _log.debug(f'add_listener: {name} {func.__name__}')
        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    def event_dispatch(
        self, event_name: str, *args: Tuple[Any], **kwargs: Dict[Any, Any]
    ) -> bool:
        """
        on_ready等といった

        Parameters
        ----------
        event_name :
        args :
        kwargs :

        Returns
        -------

        """

        ev = f'on_{event_name}'
        for event in self.special_events.get(ev, []):
            foo = importlib.import_module(event.__module__)
            coro = getattr(foo, ev)
            self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            self.schedule_event(getattr(self, ev), ev, *args, **kwargs)
        return ev in dir(self)

    def dispatch(
        self, event_name: str, *args: tuple[Any], **kwargs: Dict[Any, Any]
    ):
        ev = f'on_{event_name}'
        for event in self.extra_events.get(ev, []):
            if inspect.ismethod(event):
                coro = event
                event = event.__name__
            else:
                foo = importlib.import_module(event.__module__)
                coro = getattr(foo, ev)
            self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            self.schedule_event(getattr(self, ev), ev, *args, **kwargs)

    def schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: tuple[Any],
        **kwargs: Dict[Any, Any],
    ) -> asyncio.Task[Any]:
        return self.loop.create_task(
            self._run_event(coro, event_name, *args, **kwargs),
            name=f'MiPA: {event_name}',
        )

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                await self.__on_error(event_name)
            except asyncio.CancelledError:
                pass

    @staticmethod
    async def __on_error(event_method: str) -> None:
        print(f'Ignoring exception in {event_method}', file=sys.stderr)
        traceback.print_exc()

    async def on_error(self, err):
        self.event_dispatch('error', err)

    async def create_api_session(self, token: str, url: str) -> API:
        self.core = API(url, token)
        return self.core

    async def login(self, token: str, url: str):
        """
        ユーザーにログインし、ユーザー情報を取得します

        Parameters
        ----------
        token : str
            BOTにするユーザーのTOKEN
        url : str
            BOTにするユーザーがいるインスタンスのURL
        """

        core = await self.create_api_session(token, url)
        await core.http.login()
        self.user = await core.api.get_me()

    async def connect(
        self,
        *,
        reconnect: bool = True,
        timeout: int = 60,
        event_name: str = 'ready',
    ) -> None:
        self._connection = self._get_state()
        coro = MisskeyWebSocket.from_client(
            self, timeout=timeout, event_name=event_name
        )
        try:
            self.ws = await asyncio.wait_for(coro, timeout=60)
        except asyncio.exceptions.TimeoutError:
            await self.connect(reconnect=reconnect, timeout=timeout)

        while True:
            try:
                await self.ws.poll_event()
            except WebSocketReconnect:
                await self.connect(event_name='reconnect')

    @property
    def client(self) -> ClientActions:
        return self.core.api

    @property
    def router(self) -> Router:
        return self._router

    async def start(
        self,
        url: str,
        token: str,
        *,
        debug: bool = False,
        reconnect: bool = True,
        timeout: int = 60,
        is_ayuskey: bool = False,
        log_level: LOGING_LEVEL_TYPE = 'INFO',
    ):
        """
        Starting Bot

        Parameters
        ----------
        url: str
            Misskey Instance Websocket URL (wss://example.com)
        token: str
            User Token
        debug: bool, default False
            debugging mode
        reconnect: bool, default True
            coming soon...
        timeout: int, default 60
            Time until websocket times out
        """
        setup_logging(level=log_level)
        self.token = token
        if origin_url := re.search(r'wss?://(.*)/streaming', url):
            origin_url = (
                origin_url.group(0)
                .replace('wss', 'https')
                .replace('ws', 'http')
                .replace('/streaming', '')
            )
        else:
            origin_url = url
        self.origin_url = origin_url[:-1] if url[-1] == '/' else origin_url
        self.url = url
        await self.login(token, origin_url)
        await self.connect(reconnect=reconnect, timeout=timeout)



from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, TypeVar

import aiohttp
from mipac.util import str_lower

from mipa.exception import ClientConnectorError, WebSocketReconnect
from mipa.router import Router

if TYPE_CHECKING:
    from .client import Client

__all__ = ('MisskeyWebSocket',)


MS = TypeVar('MS', bound='aiohttp.ClientWebSocketResponse')


class MisskeyWebSocket:
    def __init__(self, socket: MS, client: Client):
        self.socket: MS = socket
        self._dispatch = lambda *args: None
        self._connection = None
        self.client = client
        self._misskey_parsers: Optional[Dict[str, Callable[..., Any]]] = None

    @classmethod
    async def from_client(
        cls, client: Client, *, timeout: int = 60, event_name: str = 'ready'
    ):
        try:
            socket = await client.core.http.session.ws_connect(
                f'{client.url}?i={client.token}'
            )
            ws = cls(socket, client)
            ws._dispatch = client.dispatch
            ws._connection = client._connection
            ws._misskey_parsers = client._connection.parsers
            client._router = Router(socket, max_capure=client.max_capture)
            client.dispatch(event_name, socket)
            return ws
        except ClientConnectorError:
            while True:
                await asyncio.sleep(3)
                return await cls.from_client(
                    client, timeout=timeout, event_name=event_name
                )

        # await ws.poll_event(timeout=timeout)

    async def received_message(self, msg, /):
        if isinstance(msg, bytes):
            msg = msg.decode()

        await self._misskey_parsers[str_lower(msg['type']).upper()](msg)

    async def poll_event(self, *, timeout: int = 60):

        msg = await self.socket.receive(timeout=timeout)

        if msg is aiohttp.http.WS_CLOSED_MESSAGE:
            await asyncio.sleep(3)
            raise WebSocketReconnect()

        elif msg.type is aiohttp.WSMsgType.TEXT:
            await self.received_message(json.loads(msg.data))



import logging
from typing import Literal

LOGING_LEVEL_TYPE = Literal[
    'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
]
LOGING_LEVELS = {
    'NOTSET': 0,
    'DEBUG': 10,
    'INFO': 20,
    'WARNING': 30,
    'ERROR': 40,
    'CRITICAL': 50,
}


def parse_logging_level(level: LOGING_LEVEL_TYPE):

    if level in LOGING_LEVELS:
        return LOGING_LEVELS[level]
    raise Exception('Not found logging level {0}' % (level))


def setup_logging(
    *,
    handler: logging.Handler | None = None,
    formatter: logging.Formatter | None = None,
    level: LOGING_LEVEL_TYPE = 'INFO',
) -> None:
    _level = parse_logging_level(level)

    if _level is None:
        _level = logging.INFO

    if handler is None:
        handler = logging.StreamHandler()

    if formatter is None:
        # if isinstance(handler, logging.StreamHandler): TODO: カラー出力に対応する
        #     pass
        # else:
        dt_fmt = '%Y-%m-%d %H:%M:%S'
        formatter = logging.Formatter(
            '[{asctime}] [{levelname:<8}] {name}: {message}', dt_fmt, style='{'
        )

    logger = logging.getLogger()

    handler.setFormatter(formatter)
    logger.setLevel(_level)
    logger.addHandler(handler)



"""Commands FrameWork用のCore部分"""

from __future__ import annotations

import asyncio
import importlib
import inspect
import re
import sys
import traceback
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from mipac.models.user import UserDetailed

from mipa import Client
from mipa.exception import (
    CogNameDuplicate,
    ExtensionAlreadyLoaded,
    ExtensionFailed,
    InvalidCogPath,
    NoEntryPointError,
)
from mipa.ext.commands.context import Context
from mipa.ext.commands.core import CommandManager

if TYPE_CHECKING:
    from aiohttp.client_ws import ClientWebSocketResponse

    from mipa.ext import Cog


__all__ = ['BotBase', 'Bot']


class BotBase(CommandManager):
    def __init__(self, **options: dict[Any, Any]):
        super().__init__(**options)
        self.extra_events: dict[str, Any] = {}
        self.special_events: dict[str, Any] = {}
        self._check_once: List[Any] = []  # TODO: いつか確認する
        self._checks: List[Any] = []  # TODO: いつか確認する
        self._after_invoke = None
        self.token: Optional[str] = None
        self.origin_uri: Optional[str] = None
        self.__extensions: dict[str, Any] = {}
        self.user: UserDetailed
        self.__cogs: dict[str, Cog] = {}
        self.strip_after_prefix = options.get('strip_after_prefix', False)
        # self.logger = get_module_logger(__name__) TODO: 直す
        self.loop = asyncio.get_event_loop()

    def _on_note(self, message):
        self.dispatch('note', message)

    async def on_ready(self, ws: ClientWebSocketResponse):
        """
        on_readyのデフォルト処理

        Parameters
        ----------
        ws : ClientWebSocketResponse
        """

    def event(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_event(func, name)
            return func

        return decorator

    def add_event(
        self, func: Coroutine[Any, Any, Any], name: Optional[str] = None
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.special_events[name].append(func)
        else:
            self.special_events[name] = [func]

    def listen(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_listener(func, name)
            return func

        return decorator

    def add_listener(
        self,
        func: Union[Coroutine[Any, Any, Any], Callable[..., Any]],
        name: Optional[str] = None,
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    async def event_dispatch(
        self, event_name: str, *args: Tuple[Any], **kwargs: dict[Any, Any]
    ) -> bool:
        """
        on_ready等といった

        Parameters
        ----------
        event_name :
        args :
        kwargs :

        Returns
        -------

        """
        ev = f'on_{event_name}'
        for event in self.special_events.get(ev, []):
            foo = importlib.import_module(event.__module__)
            coro = getattr(foo, ev)
            await self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            await self.schedule_event(getattr(self, ev), ev, *args, **kwargs)
        return ev in dir(self)

    def dispatch(
        self, event_name: str, *args: tuple[Any], **kwargs: dict[Any, Any]
    ):
        ev = f'on_{event_name}'
        for event in self.extra_events.get(ev, []):
            if inspect.ismethod(event):
                coro = event
                event = event.__name__
            else:
                foo = importlib.import_module(event.__module__)
                coro = getattr(foo, ev)
            self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            self.schedule_event(getattr(self, ev), ev, *args, **kwargs)

    def add_cog(self, cog: Cog, override: bool = False) -> None:
        cog_name = cog.__cog_name__
        existing = self.__cogs.get(cog_name)
        if existing is not None:
            if not override:
                raise CogNameDuplicate()
            self.remove_cog(cog_name)  # TODO: 作る

        cog = cog._inject(self)
        self.__cogs[cog_name] = cog

    def remove_cog(self, name: str):  # TODO: Optional[Cog]を返すように
        """Cogを削除します"""
        cog = self.__cogs.get(name)
        if cog is None:
            return

        cog._eject(self)

        return cog

    def _load_from_module(self, spec: ModuleType, key: str) -> None:
        try:
            setup = spec.setup
        except AttributeError as e:
            raise NoEntryPointError(f'{key} にsetupが存在しません') from e

        try:
            setup(self)
        except Exception as e:
            raise ExtensionFailed(key, e) from e
        else:
            self.__extensions[key] = spec

    @staticmethod
    def _resolve_name(name: str, package: Optional[str]) -> str:
        try:
            return importlib.util.resolve_name(name, package)
        except ImportError as e:
            raise InvalidCogPath(name) from e

    def load_extension(
        self, name: str, *, package: Optional[str] = None
    ) -> None:
        """拡張をロードする

        Parameters
        ----------
        name : str
            [description]
        package : Optional[str], optional
            [description], by default None
        """
        name = self._resolve_name(name, package)
        if name in self.__extensions:
            raise ExtensionAlreadyLoaded
        try:
            module = importlib.import_module(name)
        except ModuleNotFoundError as e:
            raise InvalidCogPath(f'cog: {name} へのパスが無効です') from e
        self._load_from_module(module, name)

    def schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: tuple[Any],
        **kwargs: dict[Any, Any],
    ) -> asyncio.Task[Any]:
        return asyncio.create_task(
            self._run_event(coro, event_name, *args, **kwargs),
            name=f'MiPA: {event_name}',
        )

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                await self.__on_error(event_name)
            except asyncio.CancelledError:
                pass

    @staticmethod
    async def __on_error(event_method: str) -> None:
        print(f'Ignoring exception in {event_method}', file=sys.stderr)
        traceback.print_exc()

    async def on_error(self, err):
        await self.event_dispatch('error', err)

    def get_cog(self, name: str) -> Optional[str]:
        return self.__cogs.get(name)

    async def get_context(self, message, cmd, cls=Context) -> Context:
        return cls(message=message, bot=self, cmd=cmd)

    async def progress_command(self, message):
        for cmd in self.all_commands:
            ctx = await self.get_context(message, cmd)
            if cmd.cmd_type == 'regex':
                if re.search(cmd.key, message.content):
                    hit_list = re.findall(cmd.key, message.content)
                    if isinstance(hit_list, list):
                        hit_list = tuple(hit_list)

                    if isinstance(hit_list[0], tuple):
                        hit_list = tuple(
                            i for i in hit_list[0] if len(i.rstrip()) > 0
                        )
                    ctx.args = hit_list
                    await cmd.func.invoke(ctx)
            elif message.content.find(cmd.key) != -1:
                await cmd.func.invoke(ctx)
            else:
                continue

    async def on_mention(self, message):
        await self.progress_command(message)


class Bot(BotBase, Client):
    pass



from ._version import get_versions

__title__ = 'mipa'
__author__ = 'yupix'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present yupix'
__author_email__ = 'yupi0982@outlook.jp'
__version__ = get_versions()['version']

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

del get_versions

from .client import Client
from .ext import *

__all__ = ('Client',)

from . import _version

__version__ = _version.get_versions()['version']



import asyncio
from typing import Any, Callable, Coroutine, Dict, Optional

from mipa.exception import TaskNotRunningError

__all__ = ['Loop', 'loop']


class Loop:
    def __init__(
        self,
        func: Callable[..., Coroutine[Any, Any, Any]],
        seconds: int = 60,
        custom_loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.seconds: int = seconds
        self.func: Callable[..., Coroutine[Any, Any, Any]] = func
        self._task: Optional[asyncio.Task[Any]] = None
        self.stop_next_iteration = None
        self._loop: Optional[asyncio.AbstractEventLoop] = custom_loop

    def start(
        self, *args: tuple[Any], **kwargs: Dict[Any, Any]
    ) -> asyncio.Task[Any]:
        """
        タスクを開始する

        Parameters
        ----------
        args : Any
        kwargs : Any

        Returns
        -------
        _task : asyncio.Task[Any]
        """
        _loop = (
            asyncio.get_running_loop() if self._loop is None else self._loop
        )  # self._loop が無いなら取得
        self._task = _loop.create_task(self.task(*args, **kwargs))
        return self._task

    def stop(self):
        """
        タスクを停止
        """

        if self._task is None:
            raise TaskNotRunningError('タスクは起動していません')

        if not self._task.done():
            self.stop_next_iteration = True

    async def task(self, *args: tuple[Any], **kwargs: Dict[Any, Any]):
        while True:
            if self.stop_next_iteration:
                return
            await self.func(self.seconds, *args, **kwargs)
            await asyncio.sleep(self.seconds)


def loop(n: int, custom_loop: Optional[asyncio.AbstractEventLoop] = None):
    def _deco(f: Callable[..., Coroutine[Any, Any, Any]]) -> Loop:
        return Loop(f, n, custom_loop=custom_loop)

    return _deco



from __future__ import annotations

import asyncio
import inspect
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict

from mipac.models import Note
from mipac.models.chat import ChatMessage
from mipac.models.emoji import CustomEmoji
from mipac.models.note import NoteReaction, NoteDeleted
from mipac.models.user import LiteUser
from mipac.models.reaction import PartialReaction
from mipac.types import INote
from mipac.types.chat import IChatMessage
from mipac.types.note import (
    INoteReaction,
    INoteUpdated,
    INoteUpdatedReaction,
    INoteUpdatedDelete,
)
from mipac.types.user import ILiteUser
from mipac.util import str_lower, upper_to_lower

if TYPE_CHECKING:
    from mipa.client import Client

_log = logging.getLogger(__name__)


class ConnectionState:
    def __init__(
        self,
        dispatch: Callable[..., Any],
        loop: asyncio.AbstractEventLoop,
        client: Client,
    ):
        self.__client: Client = client
        self.__dispatch = dispatch
        self.api = client.core.api
        self.loop: asyncio.AbstractEventLoop = loop
        self.parsers = parsers = {}
        for attr, func in inspect.getmembers(self):
            if attr.startswith('parse'):
                parsers[attr[6:].upper()] = func

    async def parse_emoji_added(self, message: Dict[str, Any]):
        self.__dispatch(
            'emoji_add', CustomEmoji(message['body']['emoji'], client=self.api)
        )

    async def parse_channel(self, message: Dict[str, Any]) -> None:
        """parse_channel is a function to parse channel event

        チャンネルタイプのデータを解析後適切なパーサーに移動させます

        Parameters
        ----------
        message : Dict[str, Any]
            Received message
        """
        base_msg = upper_to_lower(message['body'])
        channel_type = str_lower(base_msg.get('type'))
        _log.debug(f'ChannelType: {channel_type}')
        _log.debug(f'recv event type: {channel_type}')
        await getattr(self, f'parse_{channel_type}')(base_msg['body'])

    async def parse_renote(self, message: Dict[str, Any]):
        pass

    async def parse_unfollow(self, message: Dict[str, Any]):
        """
        フォローを解除した際のイベントを解析する関数
        """

    async def parse_signin(self, message: Dict[str, Any]):
        """
        ログインが発生した際のイベント
        """

    async def parse_receive_follow_request(self, message: Dict[str, Any]):
        """
        フォローリクエストを受け取った際のイベントを解析する関数
        """

        # self.__dispatch('follow_request', FollowRequest(message)) TODO:修正

    async def parse_note_updated(self, message: INoteUpdated[Any]):
        await getattr(self, f'parse_{message["body"]["type"]}')(
            upper_to_lower(message)
        )

    async def parse_deleted(self, note: INoteUpdated[INoteUpdatedDelete]):
        self.__dispatch('note_deleted', NoteDeleted(note))

    async def parse_unreacted(
        self, reaction: INoteUpdated[INoteUpdatedReaction]
    ):
        self.__dispatch('unreacted', PartialReaction(reaction))

    async def parse_reacted(
        self, reaction: INoteUpdated[INoteUpdatedReaction]
    ):
        self.__dispatch('reacted', PartialReaction(reaction))

    async def parse_me_updated(self, user: ILiteUser):
        self.__dispatch('me_updated', LiteUser(user, client=self.api))

    async def parse_read_all_announcements(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO: 実装

    async def parse_reply(self, message: INote) -> None:
        """
        リプライ
        """
        self.__dispatch('note', Note(message, client=self.__client.client))

    async def parse_follow(self, message: ILiteUser) -> None:
        """
        ユーザーをフォローした際のイベントを解析する関数
        """

        self.__dispatch('user_follow', LiteUser(message, client=self.api))

    async def parse_followed(self, user: ILiteUser) -> None:
        """
        フォローイベントを解析する関数
        """

        self.__dispatch('follow', LiteUser(user, client=self.api))

    async def parse_mention(self, note: INote) -> None:
        """
        メンションイベントを解析する関数
        """

        self.__dispatch('mention', Note(note, client=self.__client.client))

    async def parse_drive_file_created(self, message: Dict[str, Any]) -> None:
        self.__dispatch('drive_file_created', message)

    async def parse_read_all_unread_mentions(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    async def parse_read_all_unread_specified_notes(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    async def parse_read_all_channels(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    async def parse_read_all_notifications(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    async def parse_url_upload_finished(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    async def parse_unread_mention(self, message: Dict[str, Any]) -> None:
        pass

    async def parse_unread_specified_note(
        self, message: Dict[str, Any]
    ) -> None:
        pass

    async def parse_read_all_messaging_messages(
        self, message: Dict[str, Any]
    ) -> None:
        pass

    async def parse_messaging_message(self, message: IChatMessage) -> None:
        """
        チャットが来た際のデータを処理する関数
        """
        self.__dispatch(
            'chat', ChatMessage(message, client=self.__client.client)
        )

    async def parse_unread_messaging_message(
        self, message: IChatMessage
    ) -> None:
        """
        チャットが既読になっていない場合のデータを処理する関数
        """
        self.__dispatch(
            'chat', ChatMessage(message, client=self.__client.client)
        )

    async def parse_notification(self, message: Dict[str, Any]) -> None:
        """
        通知イベントを解析する関数

        Parameters
        ----------
        message: Dict[str, Any]
            Received message
        """

        accept_type = ['reaction']
        notification_type = str_lower(message['type'])
        if notification_type in accept_type:
            await getattr(self, f'parse_{notification_type}')(message)

    async def parse_follow_request_accepted(
        self, message: Dict[str, Any]
    ) -> None:
        pass

    async def parse_poll_vote(self, message: Dict[str, Any]) -> None:
        pass  # TODO: 実装

    async def parse_unread_notification(self, message: Dict[str, Any]) -> None:
        """
        未読の通知を解析する関数

        Parameters
        ----------
        message : Dict[str, Any]
            Received message
        """
        # notification_type = str_lower(message['type'])
        # getattr(self, f'parse_{notification_type}')(message)

    async def parse_reaction(self, message: INoteReaction) -> None:
        """
        リアクションに関する情報を解析する関数
        """
        self.__dispatch(
            'reaction', NoteReaction(message, client=self.api),
        )

    async def parse_note(self, message: INote) -> None:
        """
        ノートイベントを解析する関数
        """
        note = Note(message, self.__client.client)
        await self.__client.router.capture_message(note.id)
        self.__dispatch('note', note)
